package ssw.mj.test;

import java.io.StringReader;

import junit.framework.TestCase;

import ssw.mj.Scanner;
import ssw.mj.Token;

/**
 * Test cases for the <code>Scanner</code> class.
 *
 * @author AW
 */
public class ScannerTest extends TestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";

	private static final int
		none       = Token.none,
		ident      = Token.ident,
		number     = Token.number,
		charConst  = Token.charConst,
		plus       = Token.plus,
		minus      = Token.minus,
		times      = Token.times,
		slash      = Token.slash,
		rem        = Token.rem,
		eql        = Token.eql,
		neq        = Token.neq,
		gtr        = Token.gtr,
		geq        = Token.geq,
		less       = Token.less,
		leq        = Token.leq,
		and        = Token.and,
		or         = Token.or,
		assign     = Token.assign,
		pplus      = Token.pplus,
		mminus     = Token.mminus,
		semicolon  = Token.semicolon,
		comma      = Token.comma,
		period     = Token.period,
		lpar       = Token.lpar,
		rpar       = Token.rpar,
		lbrack     = Token.lbrack,
		rbrack     = Token.rbrack,
		lbrace     = Token.lbrace,
		rbrace     = Token.rbrace,
		classKW    = Token.classKW,
		continueKW = Token.continueKW,
		doKW       = Token.doKW,
		elseKW     = Token.elseKW,
		finalKW    = Token.finalKW,
		ifKW       = Token.ifKW,
		newKW      = Token.newKW,
		printKW    = Token.printKW,
		readKW     = Token.readKW,
		returnKW   = Token.returnKW,
		voidKW     = Token.voidKW,
		whileKW    = Token.whileKW,
		eof        = Token.eof;

	private static TestPrintWriter output;

	// ---------- helpers ----------

	private void initScanner(String s) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(s), output);
	}

	private void checkNext(int kind, int line, int col) {
		checkNext(kind, line, col, 0, null);
	}

	private void checkNext(int kind, int line, int col, String string) {
		checkNext(kind, line, col, 0, string);
	}

	private void checkNext(int kind, int line, int col, int val, String string) {
		Token expected = new Token(kind, line, col, val, string);
		Token actual = Scanner.next();
		assertEquals(expected, actual);
	}

	// ---------------------------------------
	// ---------- test valid inputs ----------

	/** Test all token types. */
	public void testAllTokens () {
		initScanner("anIdentifier 123 'c'" + LF +
					"+ - * / % == != > >= < <= && || ++ -- ; , . ( ) [ ] { }" + LF +
					"class continue do else final if new print read return void while" + LF);

		checkNext(ident, 1, 1, "anIdentifier");
		checkNext(number, 1, 14, 123, "123");
		checkNext(charConst, 1, 18, 99, null);
		checkNext(plus, 2, 1);
		checkNext(minus, 2, 3);
		checkNext(times, 2, 5);
		checkNext(slash, 2, 7);
		checkNext(rem, 2, 9);
		checkNext(eql, 2, 11);
		checkNext(neq, 2, 14);
		checkNext(gtr, 2, 17);
		checkNext(geq, 2, 19);
		checkNext(less, 2, 22);
		checkNext(leq, 2, 24);
		checkNext(and, 2, 27);
		checkNext(or, 2, 30);
		checkNext(pplus, 2, 33);
		checkNext(mminus, 2, 36);
		checkNext(semicolon, 2, 39);
		checkNext(comma, 2, 41);
		checkNext(period, 2, 43);
		checkNext(lpar, 2, 45);
		checkNext(rpar, 2, 47);
		checkNext(lbrack, 2, 49);
		checkNext(rbrack, 2, 51);
		checkNext(lbrace, 2, 53);
		checkNext(rbrace, 2, 55);
		checkNext(classKW, 3, 1, "class");
		checkNext(continueKW, 3, 7, "continue");
		checkNext(doKW, 3, 16, "do");
		checkNext(elseKW, 3, 19, "else");
		checkNext(finalKW, 3, 24, "final");
		checkNext(ifKW, 3, 30, "if");
		checkNext(newKW, 3, 33, "new");
		checkNext(printKW, 3, 37, "print");
		checkNext(readKW, 3, 43, "read");
		checkNext(returnKW, 3, 48, "return");
		checkNext(voidKW, 3, 55, "void");
		checkNext(whileKW, 3, 60, "while");
		checkNext(eof, 4, 0);

		output.verify();
	}

	/** The following short program uses CR and LF for line separation: */
	public void testCRLFLineSeparators () {
		initScanner("1" + CR + LF +
					" 2" + CR + LF +
					"  3 ");

		checkNext(number, 1, 1, 1, "1");
		checkNext(number, 2, 2, 2, "2");
		checkNext(number, 3, 3, 3, "3");
		checkNext(eof, 3, 4);
	}

	/** The following short program uses only LF for line separation: */
	public void testLFLineSeparators() {
		initScanner("1" + LF +
					" 2" + LF +
					"  3 ");

		checkNext(number, 1, 1, 1, "1");
		checkNext(number, 2, 2, 2, "2");
		checkNext(number, 3, 3, 3, "3");
		checkNext(eof, 3, 4);
	}

	public void testSimpleSingleLineComment () {
		initScanner("class SSLComment { /* This is a simple single line comment. */ }  ");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "SSLComment");
		checkNext(lbrace, 1, 18);
		checkNext(rbrace, 1, 64);
		checkNext(eof, 1, 66);
	}

	public void testSimpleMultiLineComment () {
		initScanner("class SMLComment {" + LF +
					"/*\tThis " + LF +
					"\tis " + LF +
					"\ta " + LF +
					"\tsimple " + LF +
					"\tmulti-line " + LF +
					"\t comment. */ " + LF +
					" }");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "SMLComment");
		checkNext(lbrace, 1, 18);
		checkNext(rbrace, 8, 2);
		checkNext(eof, 8, 2);
	}

	public void testNestedSingleLineComment () {
		initScanner("class NSLComment {" + LF +
					" /* This is a /* nested  /* single-line */ comment. */*/ " + LF +
					"}");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "NSLComment");
		checkNext(lbrace, 1, 18);
		checkNext(rbrace, 3, 1);
		checkNext(eof, 3, 1);
	}

	public void testNestedMultiLineComment () {
		initScanner("class NMLComment {" + LF +
					"/*" + LF +
					"\tThis is a " + LF +
					"\t/*\tnested  " + LF +
					"\t\t/* multi-line */" + LF +
					"\t\tcomment. " + LF +
					"\t*/" + LF +
					"*/ " + LF +
					"}");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "NMLComment");
		checkNext(lbrace, 1, 18);
		checkNext(rbrace, 9, 1);
		checkNext(eof, 9, 1);
	}

	public void testESCsequenceBackslash () {
		initScanner("backslash '\\\\'");

		checkNext(ident, 1, 1, "backslash");
		checkNext(charConst, 1, 11, 92, null);
		checkNext(eof, 1, 14);
	}

	public void testESCsequenceLF () {
		initScanner("linefeed '\\n'");

		checkNext(ident, 1, 1, "linefeed");
		checkNext(charConst, 1, 10, 10, null);
		checkNext(eof, 1, 13);
	}

	public void testESCsequenceCR () {
		initScanner("carriage return '\\r'");

		checkNext(ident, 1, 1, "carriage");
		checkNext(returnKW, 1, 10, "return");
		checkNext(charConst, 1, 17, 13, null);
		checkNext(eof, 1, 20);
	}

	// ----------------------------------------
	// ---------- test invalid input ----------

	public void testInvalidSymbols () {
		initScanner("ch = '';");

		checkNext(ident, 1, 1, "ch");
		checkNext(assign, 1, 4);
		checkNext(charConst, 1, 6, 0, null);
		output.addExpectedLine("-- line 1, col 7: invalid character constant ''");
		checkNext(semicolon, 1, 9);
		checkNext(eof, 1, 9);

		output.verify();

		initScanner("''");

		checkNext(charConst, 1, 1, 0, null);
		output.addExpectedLine("-- line 1, col 2: invalid character constant ''");
		checkNext(eof, 1, 3);

		output.verify();

		initScanner("");

		checkNext(none, 1, 1);
		output.addExpectedLine("-- line 1, col 1: invalid character ''");
		checkNext(eof, 1, 1);

		output.verify();
	}

	// ----------------------------------------------------
	// ---------- test complete correct programs ----------

	/** Test the shortest valid program: */
	public void testShortestProgram () {
		initScanner("class Shortest { }");

		checkNext(classKW, 1, 1, "class");
		checkNext(ident, 1, 7, "Shortest");
		checkNext(lbrace, 1, 16, null);
		checkNext(rbrace, 1, 18, null);
		checkNext(eof, 1, 18, null);
	}

	/* !!! Note: This does NOT test all productions !!! */
	public void testScriptExample () {
		initScanner("class P" + LF +
					"" + LF +
					"\tfinal int size = 10;" + LF +
					"" + LF +
		/*line  5*/	"\tclass Table {" + LF +
					"\t\tint[] pos;" + LF +
					"\t\tint[] neg;" + LF +
					"\t}" + LF +
					"" + LF +
		/*line 10*/	"\tTable val;" + LF +
					"{" + LF +
					"\tvoid main()" + LF +
					"\t\tint x, i;" + LF +
					"\t{ /*---------- Initialize val ------------*/" + LF +
		/*line 15*/	"\t\tval = new Table;" + LF +
					"\t\tval.pos = new int[size]; val.neg = new int[size];" + LF +
					"\t\ti = 0;" + LF +
					"\t\twhile (i < size) {" + LF +
					"\t\t\tval.pos[i] = 0; val.neg[i] = 0;" + LF +
		/*line 20*/	"\t\t\ti++;" + LF +
					"\t\t}" + LF +
					"\t\t/*---------- Read values ---------*/" + LF +
					"\t\tread(x);" + LF +
					"\t\twhile (x > 0) {" + LF +
		/*line 25*/	"\t\t\tif (0 <= x && x < size) {" + LF +
					"\t\t\t\tval.pos[x]++;" + LF +
					"\t\t\t} else if (-size < x && x < 0) {" + LF +
					"\t\t\t\tval.neg[-x]++;" + LF +
					"\t\t\t}" + LF +
		/*line 30*/	"\t\t\tread(x);" + LF +
					"\t\t}" + LF +
					"\t}" + LF +
					"}");

		checkNext(classKW,1,1,"class"); checkNext(ident,1,7,"P");
		checkNext(finalKW,3,2,"final"); checkNext(ident,3,8,"int"); checkNext(ident,3,12,"size"); checkNext(assign,3,17); checkNext(number,3,19,10,"10"); checkNext(semicolon,3,21);
		checkNext(classKW,5,2,"class"); checkNext(ident,5,8,"Table"); checkNext(lbrace,5,14);
		checkNext(ident,6,3,"int"); checkNext(lbrack,6,6); checkNext(rbrack,6,7); checkNext(ident,6,9,"pos"); checkNext(semicolon,6,12);
		checkNext(ident,7,3,"int"); checkNext(lbrack,7,6); checkNext(rbrack,7,7); checkNext(ident,7,9,"neg"); checkNext(semicolon,7,12);
		checkNext(rbrace,8,2);
		checkNext(ident,10,2,"Table"); checkNext(ident,10,8,"val"); checkNext(semicolon,10,11);
		checkNext(lbrace,11,1);
		checkNext(voidKW,12,2,"void"); checkNext(ident,12,7,"main"); checkNext(lpar,12,11); checkNext(rpar,12,12);
		checkNext(ident,13,3,"int"); checkNext(ident,13,7,"x"); checkNext(comma,13,8); checkNext(ident,13,10,"i"); checkNext(semicolon,13,11);
		checkNext(lbrace,14,2);
		checkNext(ident,15,3,"val"); checkNext(assign,15,7); checkNext(newKW,15,9,"new"); checkNext(ident,15,13,"Table"); checkNext(semicolon,15,18);
		checkNext(ident,16,3,"val"); checkNext(period,16,6); checkNext(ident,16,7,"pos"); checkNext(assign,16,11); checkNext(newKW,16,13, "new"); checkNext(ident,16,17,"int"); checkNext(lbrack,16,20); checkNext(ident,16,21,"size"); checkNext(rbrack,16,25); checkNext(semicolon,16,26);
		checkNext(ident,16,28,"val"); checkNext(period,16,31); checkNext(ident,16,32,"neg"); checkNext(assign,16,36); checkNext(newKW,16,38, "new"); checkNext(ident,16,42,"int"); checkNext(lbrack,16,45); checkNext(ident,16,46,"size"); checkNext(rbrack,16,50); checkNext(semicolon,16,51);
		checkNext(ident,17,3,"i"); checkNext(assign,17,5); checkNext(number,17,7,0,"0"); checkNext(semicolon,17,8);
		checkNext(whileKW,18,3,"while"); checkNext(lpar,18,9); checkNext(ident,18,10,"i"); checkNext(less,18,12); checkNext(ident,18,14,"size"); checkNext(rpar,18,18); checkNext(lbrace,18,20);
		checkNext(ident,19,4,"val"); checkNext(period,19,7); checkNext(ident,19,8,"pos"); checkNext(lbrack,19,11); checkNext(ident,19,12,"i"); checkNext(rbrack,19,13); checkNext(assign,19,15); checkNext(number,19,17,0,"0"); checkNext(semicolon,19,18);
		checkNext(ident,19,20,"val"); checkNext(period,19,23); checkNext(ident,19,24,"neg"); checkNext(lbrack,19,27); checkNext(ident,19,28,"i"); checkNext(rbrack,19,29); checkNext(assign,19,31); checkNext(number,19,33,0,"0"); checkNext(semicolon,19,34);
		checkNext(ident,20,4,"i"); checkNext(pplus,20,5); checkNext(semicolon,20,7);
		checkNext(rbrace,21,3);
		checkNext(readKW,23,3,"read"); checkNext(lpar,23,7); checkNext(ident,23,8,"x"); checkNext(rpar,23,9); checkNext(semicolon,23,10);
		checkNext(whileKW,24,3,"while"); checkNext(lpar,24,9); checkNext(ident,24,10,"x"); checkNext(gtr,24,12); checkNext(number,24,14,0,"0"); checkNext(rpar,24,15); checkNext(lbrace,24,17);
		checkNext(ifKW,25,4,"if"); checkNext(lpar,25,7); checkNext(number,25,8,0,"0"); checkNext(leq,25,10); checkNext(ident,25,13,"x"); checkNext(and,25,15); checkNext(ident,25,18,"x"); checkNext(less,25,20); checkNext(ident,25,22,"size"); checkNext(rpar,25,26); checkNext(lbrace,25,28);
		checkNext(ident,26,5,"val"); checkNext(period,26,8); checkNext(ident,26,9,"pos"); checkNext(lbrack,26,12); checkNext(ident,26,13,"x"); checkNext(rbrack,26,14); checkNext(pplus,26,15); checkNext(semicolon,26,17);
		checkNext(rbrace,27,4); checkNext(elseKW,27,6,"else"); checkNext(ifKW,27,11,"if"); checkNext(lpar,27,14); checkNext(minus,27,15); checkNext(ident,27,16,"size"); checkNext(less,27,21); checkNext(ident,27,23,"x"); checkNext(and,27,25); checkNext(ident,27,28,"x"); checkNext(less,27,30); checkNext(number,27,32,0,"0"); checkNext(rpar,27,33); checkNext(lbrace,27,35);
		checkNext(ident,28,5,"val"); checkNext(period,28,8); checkNext(ident,28,9,"neg"); checkNext(lbrack,28,12); checkNext(minus,28,13); checkNext(ident,28,14,"x"); checkNext(rbrack,28,15); checkNext(pplus,28,16); checkNext(semicolon,28,18);
		checkNext(rbrace,29,4);
		checkNext(readKW,30,4,"read"); checkNext(lpar,30,8); checkNext(ident,30,9,"x"); checkNext(rpar,30,10); checkNext(semicolon,30,11);
		checkNext(rbrace,31,3);
		checkNext(rbrace,32,2);
		checkNext(rbrace,33,1);
		checkNext(eof,33,1);

		output.verify();
	}

	public static void main (String[] args) {
		junit.textui.TestRunner.run(ScannerTest.class);
	}
}
